package master;

import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.GeometryFactory;
import org.locationtech.jts.geom.Polygon;
import org.mapdb.BTreeMap;
import org.mapdb.DB;
import org.mapdb.DBMaker;
import org.mapdb.Serializer;

import java.io.*;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

import static java.time.temporal.ChronoUnit.DAYS;


@SuppressWarnings("Duplicates")
public class Benchmark {


    BTreeMap<long[], String> map;
    FrequencyPrefixHistogram frequency_table;
    QueryExecutioner query_executioner;
    List<Query> query_set_spatial;
    List<Query> query_set_spatiotemporal;
    DB db;

    boolean best_first_traversal = false;
    boolean multithreading = false;
    boolean temporal = false;
    boolean epoch = false;

    int dimensions = 2;
    int bits_per_dimension = 31;
    int number_of_threads = 1;


    public Benchmark(List<Query> query_set_spatial, List<Query> query_set_spatiotemporal) {
        this.query_set_spatial = query_set_spatial;
        this.query_set_spatiotemporal = query_set_spatiotemporal;
        this.db = DBMaker
                .fileDB("file_test.db")
                .fileMmapEnable()
                .fileMmapEnableIfSupported()
                .fileMmapPreclearDisable()
                .cleanerHackEnable()
                .make();
    }

    public void loadDatabase(String filename) {


        try {
            if (db.isClosed()) {
                this.db = DBMaker
                        .fileDB(filename)
                        .fileMmapEnable()
                        .fileMmapEnableIfSupported()
                        .fileMmapPreclearDisable()
                        .cleanerHackEnable()
                        .make();
                this.map = db.treeMap("map")
                        .keySerializer(Serializer.LONG_ARRAY)
                        .valueSerializer(Serializer.STRING)
                        .valuesOutsideNodesEnable()
                        .createOrOpen();
            } else {
                System.out.println("Loading DB");
                db.close();
                loadDatabase(filename);
            }

        } catch (Exception e) {
            System.out.println(e.toString());
        }


    }


    public void createQueryExecutioner() {
        this.query_executioner = new QueryExecutioner(dimensions, bits_per_dimension, best_first_traversal, this.map);
        if(epoch) {
            this.query_executioner.enableEpoch('y');
        }
        if(multithreading) {
            this.query_executioner.enableMultiThreading(number_of_threads);
        }
    }

    public void setDimensions(int dimensions) {
        this.dimensions = dimensions;
    }

    public void setBitsPerDimension(int bits_per_dimension) {
        this.bits_per_dimension = bits_per_dimension;
    }

    public void enableBestFirstTraversal() {
        this.best_first_traversal = true;
        createQueryExecutioner();
    }

    public void disableBestFirstTraversal() {
        this.best_first_traversal = false;
        createQueryExecutioner();
    }

    public void createFrequencyTable(double sample_number, int depth_cut) {
        this.frequency_table = new FrequencyPrefixHistogram((int) sample_number, depth_cut, true);
        this.frequency_table.createFrequencyTable(map, dimensions, bits_per_dimension, epoch);
        this.query_executioner.setFrequencyPrefixHistogram(this.frequency_table);
    }

    public void enableSpatiotemporal() {
        this.temporal = true;
        this.dimensions = 3;
        this.bits_per_dimension = 21;
        this.epoch = false;
        loadDatabase("file_temporal_test.db");
        createQueryExecutioner();
    }

    public void enableSpatiotemporalEpoch() {
        this.temporal = true;
        this.dimensions = 3;
        this.bits_per_dimension = 21;
        this.epoch = true;
        loadDatabase("fileyear_test2.db");
        createQueryExecutioner();
    }

    public void enableSpatial() {
        this.temporal = false;
        this.dimensions = 2;
        this.bits_per_dimension = 31;
        loadDatabase("file_test.db");
        createQueryExecutioner();
    }

    public void enableMultithreading(int number_of_threads) {
        this.multithreading = true;
        this.number_of_threads = number_of_threads;
        createQueryExecutioner();
    }

    public void disableMultithreading() {
        this.multithreading = false;
        createQueryExecutioner();
    }

    public List<Query> shuffleQuerySet(List<Query> query_set) {
        Collections.shuffle(query_set);
        return query_set;
    }

    public List<List<Logger>> executeQueries(int max_ranges, boolean fetch_data) throws IOException {

        List<List<Logger>> list_loggers = new ArrayList<>();

        int number_of_iterations = 5;

        for (int i = 0; i < number_of_iterations; i++) {

            List<Logger> loggers = new ArrayList<>();

            List<Query> queries;
            if (temporal) {
                queries = shuffleQuerySet(new ArrayList<>(this.query_set_spatiotemporal));
                for (Query query : queries) {
                    Logger logger = new Logger(query.getQueryId());
                    logger.setQuery(query);
                    this.query_executioner.executeSpatiotemporalQuery(query, max_ranges, fetch_data, logger);
                    loggers.add(logger);
                }
            } else {
                queries = shuffleQuerySet(new ArrayList<>(this.query_set_spatial));
                for (Query query : queries) {
                    Logger logger = new Logger(query.getQueryId());
                    this.query_executioner.executeSpatialQuery(query, max_ranges, fetch_data, logger);
                    loggers.add(logger);

                }
            }

            list_loggers.add(loggers);
        }
        return list_loggers;
    }


    public Map<String, Double> statisticsAverageResponseTime(List<List<Logger>> list_loggers) {

        int number_of_iterations = 0;
        Map<String, Double> statistics = new HashMap<>();
        double avg_false_discovery_rate = 0;
        double avg_planning_time = 0;
        double avg_execution_time = 0;
        double avg_number_of_ranges = 0;
        int count = 0;
        for (List<Logger> loggers : list_loggers) {
            number_of_iterations++;
            double number_of_queries = 0;
            double total_false_positive_percentage = 0;
            double total_planning_time = 0;
            double total_execution_time = 0;
            double total_number_of_ranges = 0;
            for (Logger logger : loggers) {
                number_of_queries++;
                total_false_positive_percentage += logger.falseDiscoveryRate();
                total_planning_time += logger.planning_time;
                total_execution_time += logger.execution_time;
                total_number_of_ranges += logger.ranges_generated;
            }
            avg_false_discovery_rate += total_false_positive_percentage / number_of_queries;
            avg_planning_time += total_planning_time;
            avg_execution_time += total_execution_time;
            avg_number_of_ranges += total_number_of_ranges / number_of_queries;

        }


        statistics.put("AVG_FPP", avg_false_discovery_rate / number_of_iterations);
        statistics.put("AVG_PT", avg_planning_time / number_of_iterations);
        statistics.put("AVG_ET", avg_execution_time / number_of_iterations);
        statistics.put("AVG_NR", avg_number_of_ranges / number_of_iterations);

        return statistics;

    }

    public Map<String, Double> statisticsResponseTime(List<List<Logger>> list_loggers) {

        int number_of_iterations = 0;
        Map<String, Double> statistics = new HashMap<>();
        List<Integer> query_ids = new ArrayList<>();
        for (List<Logger> loggers : list_loggers) {
            number_of_iterations++;

            for (Logger logger : loggers) {
                statistics.put("FPP" + logger.logger_id, statistics.getOrDefault("FPP" + logger.logger_id, 0.0) + logger.falseDiscoveryRate());
                statistics.put("DT" + logger.logger_id, statistics.getOrDefault("DT" + logger.logger_id, 0.0) + logger.planning_time);
                statistics.put("ET" + logger.logger_id, statistics.getOrDefault("ET" + logger.logger_id, 0.0) + logger.execution_time);
                statistics.put("NR" + logger.logger_id, statistics.getOrDefault("NR" + logger.logger_id, 0.0) + logger.ranges_generated);
                statistics.put("PTET" + logger.logger_id, statistics.getOrDefault("PTET" + logger.logger_id, 0.0) + logger.planning_time + logger.execution_time);
                if (!query_ids.contains(logger.logger_id)) {
                    query_ids.add(logger.logger_id);
                    List<Coordinate> encoded_vertices = new ArrayList<>();
                    List<Coordinate> vertices = new ArrayList<>();
                    this.query_executioner.extractPolygonCoordinates(logger.query.getSpatialPredicate(), encoded_vertices, vertices);
                    GeometryFactory geometryFactory = new GeometryFactory();
                    Polygon poly = geometryFactory.createPolygon(vertices.toArray(new Coordinate[]{}));
                    statistics.put("AREA" + logger.logger_id, poly.getArea());
                    statistics.put("COUNT" + logger.logger_id, (double) logger.count);
                    statistics.put("COMPLEXITY" + logger.logger_id, Double.parseDouble(logger.query.getSpatialName().split("-")[1]));
                    statistics.put("COMPACTNESS" + logger.logger_id, Double.parseDouble(logger.query.getSpatialName().split("-")[2]));
                    statistics.put("SIZE" + logger.logger_id, Double.parseDouble(logger.query.getSpatialName().split("-")[3]));
                    if (logger.query.isTemporal()) {
                        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
                        LocalDateTime date_min = LocalDateTime.parse(logger.query.getTemporalPredicate()[0], formatter);
                        LocalDateTime date_max = LocalDateTime.parse(logger.query.getTemporalPredicate()[1], formatter);
                        statistics.put("TEMPORAL" + logger.logger_id, (double) DAYS.between(date_min, date_max));


                    }

                }
            }
        }
        int i = 0;
        for (int id : query_ids) {
            statistics.put("FPP" + id, statistics.getOrDefault("FPP" + id, 0.0) / number_of_iterations);
            statistics.put("DT" + id, statistics.getOrDefault("DT" + id, 0.0) / number_of_iterations);
            statistics.put("ET" + id, statistics.getOrDefault("ET" + id, 0.0) / number_of_iterations);
            statistics.put("NR" + id, statistics.getOrDefault("NR" + id, 0.0) / number_of_iterations);
            statistics.put("PTET" + id, statistics.getOrDefault("PTET" + id, 0.0) / number_of_iterations);
            statistics.put(Integer.toString(i)+"TEMPORAL", statistics.getOrDefault("TEMPORAL" + id, 0.0));
            statistics.put(Integer.toString(i)+"AREA", statistics.getOrDefault("AREA" + id, 0.0));
            statistics.put(Integer.toString(i)+"COUNT", statistics.getOrDefault("COUNT" + id, 0.0));
            statistics.put(Integer.toString(i)+"PTET", statistics.getOrDefault("PTET" + id, 0.0));
            statistics.put(Integer.toString(i)+"COMPLEXITY", statistics.getOrDefault("COMPLEXITY" + id, 0.0));
            statistics.put(Integer.toString(i)+"COMPACTNESS", statistics.getOrDefault("COMPACTNESS" + id, 0.0));
            statistics.put(Integer.toString(i)+"SIZE", statistics.getOrDefault("SIZE" + id, 0.0));
            statistics.put(Integer.toString(i)+"FPP", statistics.getOrDefault("FPP" + id, 0.0));
            i++;


        }
        return statistics;

    }


    public void rangeTest(int start_range, int max_ranges_limit, int increment, String filename) throws IOException {

        //disableBestFirstTraversal();


        for (int i = start_range; i < max_ranges_limit; i += increment) {

            List<List<Logger>> list_loggers = executeQueries(i, true);
            Map<String, Double> statistics = statisticsAverageResponseTime(list_loggers);
            System.out.println("Ranges: " + i + " Avg ranges created: " + statistics.get("AVG_NR") + " Avg False Positives Percentage: "
                    + statistics.get("AVG_FPP") + " Avg Planning Time: "
                    + statistics.get("AVG_PT") + " Avg Execution Time: " + statistics.get("AVG_ET")
                    + " Avg Planning + Execution Time: " + (statistics.get("AVG_PT") + statistics.get("AVG_ET")));
            String statistics_log = i + "\t" + statistics.get("AVG_NR") + "\t" + statistics.get("AVG_FPP") + "\t" + statistics.get("AVG_PT") + "\t" + statistics.get("AVG_ET") + "\t" + (statistics.get("AVG_PT") + statistics.get("AVG_ET"));
            writeStatisticsToFile(filename, statistics_log);

            if (i == 1) {
                i = 0;
            }

        }

    }

    public void frequencyTableDepthTest(double sample_number, String filename) {

        enableBestFirstTraversal();

        int root_level = 0;
        if(epoch) {
            root_level = 2;
        }

        try {
            for (int i = root_level; i < bits_per_dimension; i += 1) {

                createFrequencyTable(sample_number, i);
                List<List<Logger>> list_loggers = executeQueries(100000, true);
                Map<String, Double> statistics = statisticsAverageResponseTime(list_loggers);
                System.out.println("Depth decrease: " + i + " Avg ranges created: " + statistics.get("AVG_NR") + " Avg False Positives Percentage: "
                        + statistics.get("AVG_FPP") + " Avg Planning Time: "
                        + statistics.get("AVG_PT") + " Avg Execution Time: " + statistics.get("AVG_ET")
                        + " Avg Planning + Execution Time: " + (statistics.get("AVG_PT") + statistics.get("AVG_ET")));
                String statistics_log = i + "\t" + statistics.get("AVG_NR") + "\t" + statistics.get("AVG_FPP") + "\t" + statistics.get("AVG_PT") + "\t" + statistics.get("AVG_ET") + "\t" + (statistics.get("AVG_PT") + statistics.get("AVG_ET"));
                writeStatisticsToFile(filename, statistics_log);

            }
        } catch (Exception e) {
            System.out.println(e.toString());
        }
    }

    public void frequencyTableSampleNumberTest(int depth, String filename) {

        enableBestFirstTraversal();

        double percent = 0.001;

        try {
            for (int i = 0; i < this.map.size() * 0.10; i += this.map.size() * 0.001) {

                createFrequencyTable(i, depth);
                List<List<Logger>> list_loggers = executeQueries(100000, true);
                Map<String, Double> statistics = statisticsAverageResponseTime(list_loggers);
                System.out.println("Percent: " + percent + " Sample number: " + i + " Avg ranges created: " + statistics.get("AVG_NR") + " Avg False Positives Percentage: "
                        + statistics.get("AVG_FPP") + " Avg Planning Time: "
                        + statistics.get("AVG_PT") + " Avg Execution Time: " + statistics.get("AVG_ET")
                        + " Avg Planning + Execution Time: " + (statistics.get("AVG_PT") + statistics.get("AVG_ET")));
                String statistics_log = percent + "\t" + i + "\t" + statistics.get("AVG_NR") + "\t" + statistics.get("AVG_FPP") + "\t" + statistics.get("AVG_PT") + "\t" + statistics.get("AVG_ET") + "\t" + (statistics.get("AVG_PT") + statistics.get("AVG_ET"));
                writeStatisticsToFile(filename, statistics_log);
                percent += 0.001;

            }
        } catch (Exception e) {
            System.out.println(e.toString());
        }


    }

    public void executeQueriesDetailedStatistics(int depth, double sample_number, int maximum_ranges) throws IOException, InterruptedException {
        ////// Options//////
        enableBestFirstTraversal();
        createFrequencyTable(sample_number, depth);
        //Thread.sleep(10000000);

        ////////////////////
        List<List<Logger>> list_loggers = executeQueries(maximum_ranges, true);
        Map<String, Double> statistics = statisticsResponseTime(list_loggers);
        String response_time = "";
        String decomposistion_time = "";
        for (String key : statistics.keySet()) {
            if (key.startsWith("PTET")) {
                response_time += statistics.get(key) + "\\\\  ";
            }
        }
        for (String key : statistics.keySet()) {
            if (key.startsWith("DT")) {
                decomposistion_time += statistics.get(key) + "\\\\  ";
            }
        }

        System.out.println("Response time:");
        System.out.println(response_time);
        System.out.println("Decompositoin time");
        System.out.println(decomposistion_time);


        String response_time_by_area = "";
        String response_time_by_count = "";
        String count_by_area = "";
        String count_by_temporal_width = "";
        String response_time_by_temporal_day = "";
        String response_time_by_temporal_week = "";
        String response_time_by_temporal_month = "";
        String average_false_detection_rate_by_complexity = "";
        String average_false_detection_rate_by_compactness = "";
        ArrayList<Point> reponse_time_by_area_list = new ArrayList<>();
        ArrayList<Point> reponse_time_by_count_list = new ArrayList<>();
        ArrayList<Point> count_by_area_list = new ArrayList<>();
        ArrayList<Point> count_by_temporal_width_list = new ArrayList<>();
        ArrayList<Point> response_time_by_complexity_list = new ArrayList<>();
        ArrayList<Point> response_time_by_compactness_list = new ArrayList<>();
        ArrayList<Point> average_false_detection_rate_by_complexity_list = new ArrayList<>();
        ArrayList<Point> average_false_detection_rate_by_compactness_list = new ArrayList<>();
        int query_size;
        if(this.temporal) {
            query_size = this.query_set_spatiotemporal.size();
        }
        else {
            query_size = this.query_set_spatial.size();
        }

        System.out.println("QUERY SIZE: " + query_size);

        for(int i = 0; i<query_size; i++){


            reponse_time_by_area_list.add(new Point(statistics.get(Integer.toString(i)+"AREA"), statistics.get(Integer.toString(i)+"PTET"), 0));
            reponse_time_by_count_list.add(new Point(statistics.get(Integer.toString(i)+"COUNT"), statistics.get(Integer.toString(i)+"PTET"), 0));
            count_by_area_list.add(new Point(statistics.get(Integer.toString(i)+"AREA"), statistics.get(Integer.toString(i)+"COUNT"), 0));
            response_time_by_complexity_list.add(new Point(statistics.get(Integer.toString(i)+"COMPLEXITY"), statistics.get(Integer.toString(i)+"PTET"), statistics.get(Integer.toString(i)+"SIZE").intValue()));
            response_time_by_compactness_list.add(new Point(statistics.get(Integer.toString(i)+"COMPACTNESS"), statistics.get(Integer.toString(i)+"PTET"), statistics.get(Integer.toString(i)+"SIZE").intValue()));
            average_false_detection_rate_by_complexity_list.add(new Point(statistics.get(Integer.toString(i)+"COMPLEXITY"), statistics.get(Integer.toString(i)+"FPP"), 0));
            average_false_detection_rate_by_compactness_list.add(new Point(statistics.get(Integer.toString(i)+"COMPACTNESS"), statistics.get(Integer.toString(i)+"FPP"), 0));

            int temporal_width = statistics.get(Integer.toString(i) + "TEMPORAL").intValue();
            if(temporal) {
                if (temporal_width == 1) {
                    count_by_temporal_width_list.add(new Point(1, statistics.get(Integer.toString(i)+"COUNT"), 0));
                    response_time_by_temporal_day += statistics.get(Integer.toString(i) + "PTET") + "\\\\  ";
                } else if (temporal_width == 7) {
                    count_by_temporal_width_list.add(new Point(2, statistics.get(Integer.toString(i)+"COUNT"), 0));
                    response_time_by_temporal_week += statistics.get(Integer.toString(i) + "PTET") + "\\\\  ";
                } else {
                    count_by_temporal_width_list.add(new Point(3, statistics.get(Integer.toString(i)+"COUNT"), 0));
                    response_time_by_temporal_month += statistics.get(Integer.toString(i) + "PTET") + "\\\\  ";
                }
            }
        }

        Collections.sort(reponse_time_by_area_list);
        Collections.sort(reponse_time_by_count_list);
        Collections.sort(count_by_area_list);
        Collections.sort(response_time_by_complexity_list);
        Collections.sort(response_time_by_compactness_list);
        Collections.sort(average_false_detection_rate_by_complexity_list);
        Collections.sort(average_false_detection_rate_by_compactness_list);
        for(Point p : reponse_time_by_area_list) {
            response_time_by_area += p.toString();
        }
        for(Point p : reponse_time_by_count_list) {
            response_time_by_count += p.toString();
        }
        for(Point p : count_by_area_list) {
            count_by_area += p.toString();
        }
        for(Point p : count_by_temporal_width_list) {
            count_by_temporal_width += p.toString();
        }

        for(Point p : average_false_detection_rate_by_complexity_list) {
            average_false_detection_rate_by_complexity += p.toString();
        }
        for(Point p : average_false_detection_rate_by_compactness_list) {
            average_false_detection_rate_by_compactness += p.toString();
        }

        List<String> size_complexity = Arrays.asList("", "", "", "", "", "");
        List<String> size_compactness = Arrays.asList("", "", "", "", "", "");

        for(Point p : response_time_by_complexity_list) {
            size_complexity.set(p.z, size_complexity.get(p.z) + p.toString());
        }
        for(Point p : response_time_by_compactness_list) {
            size_compactness.set(p.z, size_compactness.get(p.z) + p.toString());
        }





        System.out.println("Response time by area:");
        System.out.println(response_time_by_area);
        System.out.println("Response time by count:");
        System.out.println(response_time_by_count);
        System.out.println("Response time by day:");
        System.out.println(response_time_by_temporal_day);
        System.out.println("Response time by week:");
        System.out.println(response_time_by_temporal_week);
        System.out.println("Response time by month:");
        System.out.println(response_time_by_temporal_month);
        System.out.println("Query count by area:");
        System.out.println(count_by_area);
        System.out.println("Query count by temporal width:");
        System.out.println(count_by_temporal_width);
        for(int i = 1; i<6; i++) {
            System.out.println("Response time by complexity and size: " + i);
            System.out.println(size_complexity.get(i));
        }

        System.out.println("");

        for(int i = 1; i<6; i++) {
            System.out.println("Response time by compactness and size: " + i);
            System.out.println(size_compactness.get(i));
        }

        System.out.println("False discovery rate by complexity: ");
        System.out.println(average_false_detection_rate_by_complexity);
        System.out.println("False discovery rate by compactness");
        System.out.println(average_false_detection_rate_by_compactness);






    }

    public void runBenchmark() throws IOException, InterruptedException {


        //enableMultithreading(4);


        readStatistics();
        //enableSpatiotemporal();





        //disableBestFirstTraversal();

        enableSpatiotemporalEpoch();

        executeQueriesDetailedStatistics(3, map.size() * 0.02, 100000);


        db.close();
        System.exit(0);

    }


    public void writeStatisticsToFile(String filename, String output) {

        try (PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filename, true)))) {
            out.println(output + "\n");
        } catch (IOException e) {
            System.err.println(e);
        }

    }


    public void readStatistics() throws IOException {

        String statistics = "C:\\users\\Vigleik\\Desktop\\Masteroppgave\\Kode-java\\sample.txt";
        File statistics_f = new File(statistics);


        List<String> ranges = new ArrayList<>();
        List<String> AVG_NR = new ArrayList<>();
        List<String> AVG_FPP = new ArrayList<>();
        List<String> AVG_PT = new ArrayList<>();
        List<String> AVG_ET = new ArrayList<>();
        List<String> AVG_PT_ET = new ArrayList<>();
        List<String> PERCENT = new ArrayList<>();



        try (Scanner in = new Scanner(new FileReader(statistics));) {
            String line;
            in.useDelimiter("\n");
            while (in.hasNext()){


                String[] parts = in.next().split("\t");

                if(in.hasNext()) {
                    in.next();

                    //PERCENT.add(parts[0]);
                    ranges.add(parts[0]);
                    AVG_NR.add(parts[1]);
                    AVG_FPP.add(parts[2]);
                    AVG_PT.add(parts[3]);
                    AVG_ET.add(parts[4]);
                    // double resp = Double.parseDouble(parts[4]) + Double.parseDouble(parts[5]);
                    AVG_PT_ET.add(parts[6]);
                }

            }

        }

        String str = "";

        for(int i = 0; i < ranges.size(); i++) {

            str += "(" + (Double.parseDouble(ranges.get(i)))*100 + "," + Double.parseDouble(AVG_ET.get(i))/450 + ")";
        }

        System.out.println(str);


    }


}










