package master;

import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.GeometryFactory;
import org.locationtech.jts.geom.Polygon;
import org.locationtech.jts.geom.prep.PreparedGeometry;
import org.locationtech.jts.geom.prep.PreparedPolygon;

import java.util.*;
import java.util.concurrent.Callable;

public class Decomposition implements Callable<List<ZorderRange>> {

    private final int number_of_dimensions;
    private final int bits_per_dimensions;
    private final int z_address_length;
    private final int branching_factor;
    private List<ZorderRange> ranges;
    private final long restrict_precision;
    private QueryRegion query_region;
    private FrequencyPrefixHistogram frequency_prefix_histogram_map;
    private long offset;
    private PriorityQueue<UnprocessedHyperQuadrant> priority_queue;
    private Queue<UnprocessedHyperQuadrant> queue;
    private long timestamp;
    private int max_ranges;
    private int thread_id;
    private boolean best_first_traversal;
    private double frequency_threshold;
    private PreparedGeometry prepared_query_region_polygon;
    private boolean enable_epoch;


    private final ZorderCurve m;


    private final GeometryFactory geometry_factory = new GeometryFactory();

    public Decomposition(int number_of_dimensions, int bits_per_dimension, boolean enable_epoch)

    {

        this.number_of_dimensions = number_of_dimensions;
        this.bits_per_dimensions = bits_per_dimension;
        this.z_address_length = number_of_dimensions * bits_per_dimensions;
        this.branching_factor = (int) Math.pow(2, this.number_of_dimensions);
        this.m = new ZorderCurve(number_of_dimensions);
        this.restrict_precision = 64;
        this.enable_epoch = enable_epoch;

    }


    public void setDecompositionParameters(QueryRegion query_region, int max_ranges, FrequencyPrefixHistogram FrequencyPrefixHistogram, int thread_id, boolean best_first_traversal, double frequency_threshold) {

        this.query_region = query_region;
        this.frequency_prefix_histogram_map = FrequencyPrefixHistogram;
        this.timestamp = 0;
        this.frequency_threshold = frequency_threshold;
        this.max_ranges = max_ranges;
        this.ranges = new ArrayList<>();
        this.thread_id = thread_id;
        this.best_first_traversal = best_first_traversal;
        if (best_first_traversal) {
            this.priority_queue = new PriorityQueue<>(max_ranges);
        } else {
            this.queue = new LinkedList<>();
        }
        this.prepared_query_region_polygon = new PreparedPolygon(this.query_region.polygon);
        if (!this.query_region.polygon.isValid()) {
            System.out.println("GEOMETRY IS INVALID");
            System.exit(0);
        }
    }

    @Override
    public List<ZorderRange> call() {
        return createRanges(false);
    }


    private boolean contained(long min_temporal_offset, long max_temporal_offset, Polygon quadrant) {

        if (number_of_dimensions == 3) {

            return min_temporal_offset >= this.query_region.temporal_offset_min && max_temporal_offset <= this.query_region.temporal_offset_max && this.prepared_query_region_polygon.contains(quadrant);

        } else {
            return prepared_query_region_polygon.contains(quadrant); //
        }

    }

    private boolean intersects(long min_temporal_offset, long max_temporal_offset, Polygon quadrant) {

        if (number_of_dimensions == 3) {

            return ((max_temporal_offset <= this.query_region.temporal_offset_max && max_temporal_offset >= this.query_region.temporal_offset_min) || (min_temporal_offset >= this.query_region.temporal_offset_min && min_temporal_offset <= this.query_region.temporal_offset_max) ||
                    max_temporal_offset >= this.query_region.temporal_offset_max && min_temporal_offset <= this.query_region.temporal_offset_min) && this.prepared_query_region_polygon.intersects(quadrant);
        } else {
            return prepared_query_region_polygon.intersects(quadrant);

        }
    }


    private Polygon getHyperQuadrant(long[] min_decoded, long[] max_decoded) {

        Coordinate[] coordinates = new Coordinate[]
                {new Coordinate(min_decoded[0], min_decoded[1]), new Coordinate(min_decoded[0], max_decoded[1]),
                        new Coordinate(max_decoded[0], max_decoded[1]), new Coordinate(max_decoded[0], min_decoded[1]),
                        new Coordinate(min_decoded[0], min_decoded[1])};

        return geometry_factory.createPolygon(coordinates);
    }


    public long[] getLongestCommonPrefix(long[] z_addresses) {


        long shift = z_address_length - number_of_dimensions;
        long prefix = z_addresses[0] >>> shift;

        boolean processing = true;

        while (processing) {
            for (long val : Arrays.copyOfRange(z_addresses, 1, z_addresses.length)) {
                if (!(((val >>> shift) == prefix) && shift > -1)) {
                    processing = false;
                    shift += number_of_dimensions;
                    break;
                }
            }
            if (processing) {
                shift -= number_of_dimensions;
                prefix = z_addresses[0] >>> shift;
            }
        }


        return new long[]{z_addresses[0] & (Long.MAX_VALUE << shift), restrict_precision - shift};

    }

    public void setOffset(long offset) {
        this.offset = offset;
    }


    private double getPriorityValue(long quadrant_prefix) {
        double priority_value = 0;

        if (quadrant_prefix != 0) {
            long key;
            if (number_of_dimensions == 3 && enable_epoch) {
                key = Long.parseLong(Long.toBinaryString(query_region.epoch << (this.z_address_length - Long.toBinaryString(quadrant_prefix).length())) + Long.toBinaryString(quadrant_prefix >> frequency_prefix_histogram_map.bit_precision), 2);

                if (frequency_prefix_histogram_map.epoch_frequency_histogram_map.containsKey(key)) {
                    priority_value = frequency_prefix_histogram_map.epoch_frequency_histogram_map.get(key);
                }
            } else {
                key = quadrant_prefix;
                if (frequency_prefix_histogram_map.frequency_prefix_histogram_map.containsKey(key)) {
                    priority_value = frequency_prefix_histogram_map.frequency_prefix_histogram_map.get(key);
                }
            }

        }

        return priority_value;
    }


    public void checkHyperQuadrant(long hyper_quadrant_prefix, long sub_quadrant, long level) {

        long hyper_quadrant_offset;

        if (level == -1) {
            hyper_quadrant_offset = this.offset;
        } else {
            hyper_quadrant_offset = this.offset - level * this.number_of_dimensions;
        }

        long zbound_lower = hyper_quadrant_prefix | (sub_quadrant << hyper_quadrant_offset);
        long zbound_upper = zbound_lower | (1L << hyper_quadrant_offset) - 1;

        long min_temporal_offset = 0;
        long max_temporal_offset = 0;
        long[] min_decoded = m.decodeZaddress(zbound_lower);
        long[] max_decoded = m.decodeZaddress(zbound_upper);
        Polygon quadrant_polygon = getHyperQuadrant(min_decoded, max_decoded);
        if (number_of_dimensions == 3) {
            min_temporal_offset = min_decoded[2];
            max_temporal_offset = max_decoded[2];
        }

        if (contained(min_temporal_offset, max_temporal_offset, quadrant_polygon) || hyper_quadrant_offset < (64 - this.restrict_precision)) {
            ranges.add(new ZorderRange(zbound_lower, zbound_upper, true));
        } else if (intersects(min_temporal_offset, max_temporal_offset, quadrant_polygon)) {
            this.timestamp++;
            double priority_value = 0;
            if (best_first_traversal) {
                if (hyper_quadrant_offset >= frequency_prefix_histogram_map.bit_precision) {
                    priority_value = getPriorityValue(zbound_lower);
                }
                priority_queue.add(new UnprocessedHyperQuadrant(new long[]{zbound_lower, zbound_upper}, level + 1, priority_value, this.timestamp));

            } else {
                queue.add(new UnprocessedHyperQuadrant(new long[]{zbound_lower, zbound_upper}, level + 1, priority_value, this.timestamp));
            }

        }

    }


    public List<ZorderRange> createRanges(boolean initial_search) {


        boolean cont = true;

        while (cont) {

            UnprocessedHyperQuadrant next_hyper_quadrant;
            if (best_first_traversal) {
                next_hyper_quadrant = priority_queue.poll();
            } else {
                next_hyper_quadrant = queue.poll();
            }

            for (long sub_hyper_quadrant = 0L; sub_hyper_quadrant < branching_factor; sub_hyper_quadrant++) {

                checkHyperQuadrant(next_hyper_quadrant.z_prefix_range[0], sub_hyper_quadrant, next_hyper_quadrant.level);
            }

            if (best_first_traversal) {

                cont = !priority_queue.isEmpty() && (next_hyper_quadrant.estimated_objects_contained > this.frequency_threshold || ranges.size() < 1) && (ranges.size() + priority_queue.size()) < max_ranges;
            } else {

                cont = !queue.isEmpty() && (ranges.size() + queue.size()) < max_ranges;

            }

        }

        Collection<UnprocessedHyperQuadrant> unprocessed_hyper_quadrants;
        if (best_first_traversal) {
            unprocessed_hyper_quadrants = new ArrayList<>(priority_queue);
        } else {
            unprocessed_hyper_quadrants = new ArrayList<>(queue);

        }


        if (!initial_search) {
            for (UnprocessedHyperQuadrant unprocessed_hyper_quadrant : unprocessed_hyper_quadrants) {
                long[] z_prefix_range = unprocessed_hyper_quadrant.z_prefix_range;
                ranges.add(new ZorderRange(z_prefix_range[0], z_prefix_range[1], false));
            }
        }

        return ranges;

    }

    public Collection<UnprocessedHyperQuadrant> getUnprocessedHyperQuadrants() {

        if (best_first_traversal) {
            return new ArrayList<>(this.priority_queue);
        } else {
            return new ArrayList<>(this.queue);
        }

    }

    public void setUnprocessedHyperQuadrants(Collection<UnprocessedHyperQuadrant> remaining) {

        if (best_first_traversal) {
            this.priority_queue = new PriorityQueue<>(remaining);
        } else {
            this.queue = new LinkedList<>(remaining);
        }
    }


}
