package master;

import org.mapdb.BTreeMap;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;

import static java.lang.Math.pow;

public class Loader {


    String root_directory;
    Long seqNumber;
    int dimensions;
    int bits_per_dimensions;
    ZorderCurve m;
    char period;
    boolean enable_epoch;


    public Loader(String root_directory, int dimensions, int bits_per_dimensions) {
        this.root_directory = root_directory;
        this.seqNumber = 0L;
        this.dimensions = dimensions;
        this.bits_per_dimensions = bits_per_dimensions;
        this.m = new ZorderCurve(dimensions);
        this.enable_epoch = false;

    }

    public void enableEpoch(char period) {
        this.period = period;
        this.enable_epoch = true;
    }

    public long[] encodeFeatureZ2Key(String lat, String lon) {


        long scaled_lat = ((long) ((Double.parseDouble(lat) + 90) / 180 * pow(2, this.bits_per_dimensions)));
        long scaled_lon = ((long) ((Double.parseDouble(lon) + 180) / 360 * pow(2, this.bits_per_dimensions)));

        long z2key;
        long[] key;
        try {
            z2key = m.encodeZaddress(new long[]{scaled_lat, scaled_lon});
            key = new long[]{z2key, this.seqNumber};
        }
        catch(Morton64Exception exp) {
            System.out.println("Invalid coordinates: " + lat + ", " + lon);
            key = null;
        }

        return key;
    }

    public long[] encodeFeatureZ3Key(String lat, String lon, long epoch, long offset) {


        long scaled_lat = ((long) ((Double.parseDouble(lat) + 90) / 180 * pow(2, this.bits_per_dimensions)));
        long scaled_lon = ((long) ((Double.parseDouble(lon) + 180) / 360 * pow(2, this.bits_per_dimensions)));
        long scaled_offset;
        long[] key;
        if (this.enable_epoch) {
            if (this.period == 'w') {
                scaled_offset = (long) (((double) offset / 604800 * pow(2, this.bits_per_dimensions)));
            } else if (this.period == 'm') {
                scaled_offset = (long) (((double) offset / 2678400 * pow(2, this.bits_per_dimensions)));
            } else {
                scaled_offset = (long) (((double) offset / 527040 * pow(2, this.bits_per_dimensions)));
            }
            try {
                long z3key = m.encodeZaddress(new long[]{scaled_lat, scaled_lon, scaled_offset});
                key = new long[]{epoch, z3key, this.seqNumber};
            }
            catch(Morton64Exception exp) {
                System.out.println("Invalid coordinates: " + lat + ", " + lon);
                key = null;
            }

        } else {
            scaled_offset = (long) (((double) offset / 2720000 * pow(2, this.bits_per_dimensions)));
            long z3key;
            try {
                z3key = m.encodeZaddress(new long[]{scaled_lat, scaled_lon, scaled_offset});
                key = new long[]{z3key, this.seqNumber};
            }
            catch(Morton64Exception exp) {
                System.out.println("Invalid coordinates: " + lat + ", " + lon);
                key = null;
            }


        }
        return key;
    }

    public void loadData(BTreeMap<long[], String> map) throws IOException {


        File parent = new File(this.root_directory);
        for (File subFolder : parent.listFiles()) {
            if (subFolder.isDirectory()) {
                File user = new File(subFolder.getPath() + "\\Trajectory");
                String userId = subFolder.getName();
                for (File traj : user.listFiles()) {
                    String trackId = traj.getName().substring(0, traj.getName().length() - 4);
                    try (BufferedReader br = new BufferedReader(new FileReader(traj))) {
                        String line;
                        for (int i = 1; i <= 6; i++) {
                            br.readLine();
                        }
                        while ((line = br.readLine()) != null) {
                            String[] feature = line.split(",");
                            String lat = feature[0];
                            String lon = feature[1];

                            long[] key;

                            if (this.dimensions == 2) {
                                key = encodeFeatureZ2Key(lat, lon);
                            } else {
                                DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
                                LocalDateTime date = LocalDateTime.parse(feature[5] + " " + feature[6], formatter);
                                long offset;
                                long epoch;
                                if (this.enable_epoch) {
                                    if (this.period == 'w') {
                                        epoch = ChronoUnit.WEEKS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date);
                                        LocalDateTime epoch_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusWeeks(epoch);
                                        offset = ChronoUnit.SECONDS.between(epoch_date, date);
                                    } else if (this.period == 'm') {
                                        epoch = ChronoUnit.MONTHS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date);
                                        LocalDateTime epoch_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusMonths(epoch);
                                        offset = ChronoUnit.SECONDS.between(epoch_date, date);
                                    } else {
                                        epoch = ChronoUnit.YEARS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date);
                                        LocalDateTime epoch_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusYears(epoch);
                                        offset = ChronoUnit.MINUTES.between(epoch_date, date);
                                    }

                                    key = encodeFeatureZ3Key(lat, lon, epoch, offset);
                                } else {
                                    LocalDateTime epoch_start = LocalDateTime.parse("2007-04-01 00:00:00", formatter);
                                    offset = ChronoUnit.MINUTES.between(epoch_start, date);
                                    key = encodeFeatureZ3Key(lat, lon, 0, offset);
                                }
                            }
                            if(key != null) {
                                map.put(key, line + "," + userId + "," + trackId);
                                this.seqNumber++;
                            }

                        }
                    }
                }
            }
        }


    }

    public List<Query> loadQueries(boolean temporal) throws IOException {

        String spatial__query_file = "C:\\users\\Vigleik\\Desktop\\Masteroppgave\\spatial_predicates_spatial.txt";

        if(temporal) {
            spatial__query_file = "C:\\users\\Vigleik\\Desktop\\Masteroppgave\\spatial_predicates_spatiotemporal.txt";
        }

        List<Query> queries = new ArrayList<>();
        File spatial_queries = new File(spatial__query_file);
        List<List<String[]>> spatial_predicates = new ArrayList<>();
        List<String> spatial_name = new ArrayList<>();
        try (BufferedReader br = new BufferedReader(new FileReader(spatial_queries))) {
            String line;
            br.readLine();
            while ((line = br.readLine()) != null) {

                String[] parts = line.split("\t");
                String poly = parts[4].replaceAll("[()]", "");
                String coordinates[] = poly.split(",");

                List<String[]> spatial_predicate = new ArrayList<>();
                for (String coord : coordinates) {
                    String[] lat_lon = coord.trim().split("\\s+");
                    spatial_predicate.add(new String[]{lat_lon[1], lat_lon[0]});

                }
                spatial_predicates.add(spatial_predicate);
                spatial_name.add(parts[0] + "-" + parts[1] + "-" + parts[2] + "-" + parts[3]);
            }

        }


        int query_id = 0;


        if (temporal) {
            String temporal_quuery_file = "C:\\users\\Vigleik\\Desktop\\Masteroppgave\\temporal_predicates_spatiotemporal.txt";
            File temporal_queries = new File(temporal_quuery_file);
            List<String[]> temporal_predicates = new ArrayList<>();
            try (BufferedReader br = new BufferedReader(new FileReader(temporal_queries))) {
                String line;
                br.readLine();
                while ((line = br.readLine()) != null) {

                    String[] parts = line.split("\t");
                    String time_range[] = new String[]{parts[0], parts[1], parts[2].trim(), parts[3].trim()};
                    temporal_predicates.add(time_range);

                }

            }
            for (int i = 0; i < spatial_predicates.size(); i++) {
                for (String[] temporal_predicate : temporal_predicates) {

                    if(spatial_name.get(i).split("-")[0].equalsIgnoreCase(temporal_predicate[1])) {
                        queries.add(new Query(spatial_name.get(i), spatial_predicates.get(i), new String[]{temporal_predicate[2], temporal_predicate[3]}, query_id));
                        query_id++;
                    }

                }
            }
        } else {
            for (int i = 0; i < spatial_predicates.size(); i++) {

                queries.add(new Query(spatial_name.get(i), spatial_predicates.get(i), null, query_id));
                query_id++;

            }
        }

        return queries;
    }


}
