package master;

import com.google.common.collect.Iterables;
import org.locationtech.jts.geom.Coordinate;
import org.locationtech.jts.geom.GeometryFactory;
import org.locationtech.jts.geom.Polygon;
import org.mapdb.BTreeMap;
import org.openjdk.jmh.annotations.Benchmark;

import java.io.IOException;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.*;

import static java.lang.Math.pow;

public class QueryExecutioner {

    private boolean multithreading;
    private int cores;
    private ExecutorService service;

    private final int bits_per_dimensions;
    private char period;
    private final GeometryFactory geometry_factory;
    private final QueryPlanner query_planner;
    private final BTreeMap<long[], String> map;
    private final boolean best_first_traversal;
    private boolean enable_epoch;


    public QueryExecutioner(int dimensions, int bits_per_dimensions, boolean best_first_traversal, BTreeMap<long[], String> map) {
        this.bits_per_dimensions = bits_per_dimensions;
        this.geometry_factory = new GeometryFactory();
        this.map = map;
        this.query_planner = new QueryPlanner(dimensions, this.bits_per_dimensions);
        this.best_first_traversal = best_first_traversal;
        this.enable_epoch = false;
    }

    public void enableMultiThreading(int cores) {
        this.multithreading = true;
        this.query_planner.enableMultithreading(cores);
        this.cores = cores;

    }

    public void enableEpoch(char period) {
        this.period = period;
        this.enable_epoch = true;
    }

    public void setFrequencyPrefixHistogram(FrequencyPrefixHistogram FrequencyPrefixHistogram) {
        this.query_planner.setFrequencyPrefixHistogram(FrequencyPrefixHistogram);
    }

    public Iterable<String> executeSpatialQuery(Query query, int max_ranges, boolean fetch_data, Logger logger) {

        logger.setQuery(query);
        logger.startParsing();


        List<Coordinate> encoded_vertices = new ArrayList<>();
        List<Coordinate> vertices = new ArrayList<>();

        extractPolygonCoordinates(query.getSpatialPredicate(), encoded_vertices, vertices);

        Polygon encoded_poly = geometry_factory.createPolygon(encoded_vertices.toArray(new Coordinate[]{}));
        Polygon poly = geometry_factory.createPolygon(vertices.toArray(new Coordinate[]{}));

        List<Collection<String>> features = new ArrayList<>();

        QueryRegion bound = new QueryRegion(this.bits_per_dimensions, encoded_poly);

        logger.endParsing();
        logger.startCreateRanges();

        List<ZorderRange> ranges = query_planner.regionDecompostion(bound, max_ranges, best_first_traversal);


        logger.endCreateRanges(ranges.size());
        logger.startExecution();


        long false_positive_count = 0;

        if (this.multithreading) {
            this.service = Executors.newFixedThreadPool(this.cores);
        }

        List<Callable<RefinedFeatures>> tasks = new ArrayList<>();

        for (ZorderRange range : ranges) {

            ConcurrentNavigableMap<long[], String> features_retrieved =
                    map.subMap(
                            new long[]{range.lower_range}, true,
                            new long[]{range.upper_range}, true
                    );


            if (range.is_contained) {

                features.add(features_retrieved.values());

            } else {


                if (this.multithreading) {

                    Filter filter = new Filter(features_retrieved);
                    filter.setSpatialPredicate(poly);
                    tasks.add(filter);

                } else {

                    Filter filter = new Filter(features_retrieved);
                    filter.setSpatialPredicate(poly);
                    RefinedFeatures filtered_features = filter.call();
                    if (!filtered_features.getRefinedFeatures().isEmpty()) {
                        if (fetch_data) {
                            logger.startFetching();
                            features.add((filtered_features.getRefinedFeatures()));
                            logger.endFetching();
                        }
                        false_positive_count += filtered_features.getFalsePositiveCount();
                    }

                }

            }

        }

        long[] false_positive_and_refined_count = multithreadFilter(features, tasks, fetch_data, logger);
        false_positive_count += false_positive_and_refined_count[0];

        Iterable<String> results = Iterables.concat(features);

        long count = 0;

        if (fetch_data) {
            logger.startFetching();
            count = fetchData(results);
            logger.endFetching();
        }

        logger.endExecution(count, false_positive_count);

        return results;
    }


    public Iterable<String> executeSpatiotemporalQuery(Query query, int max_ranges, boolean fetch_data, Logger logger) throws IOException {
        logger.setQuery(query);
        logger.startParsing();

        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
        LocalDateTime date_min = LocalDateTime.parse(query.getTemporalPredicate()[0], formatter);
        LocalDateTime date_max = LocalDateTime.parse(query.getTemporalPredicate()[1], formatter);
        long epoch_min = 0;
        long epoch_max = 0;
        long scaled_offset_epoch_min;
        long scaled_offset_epoch_max;
        if (enable_epoch) {
            switch (period) {
                case 'w': {
                    epoch_min = ChronoUnit.WEEKS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_min);
                    epoch_max = ChronoUnit.WEEKS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_max);
                    LocalDateTime epoch_min_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusWeeks(epoch_min);
                    LocalDateTime epoch_max_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusWeeks(epoch_max);
                    long offset_min = ChronoUnit.SECONDS.between(epoch_min_date, date_min);
                    long offset_max = ChronoUnit.SECONDS.between(epoch_max_date, date_max);
                    scaled_offset_epoch_min = (long) (((double) offset_min / 604800) * pow(2, this.bits_per_dimensions));
                    scaled_offset_epoch_max = (long) (((double) offset_max / 604800) * pow(2, this.bits_per_dimensions));
                    break;
                }
                case 'm': {
                    epoch_min = ChronoUnit.MONTHS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_min);
                    epoch_max = ChronoUnit.MONTHS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_max);
                    LocalDateTime epoch_min_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusMonths(epoch_min);
                    LocalDateTime epoch_max_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusMonths(epoch_max);
                    long offset_min = ChronoUnit.SECONDS.between(epoch_min_date, date_min);
                    long offset_max = ChronoUnit.SECONDS.between(epoch_max_date, date_max);
                    scaled_offset_epoch_min = (long) (((double) offset_min / 2678400) * pow(2, this.bits_per_dimensions));
                    scaled_offset_epoch_max = (long) (((double) offset_max / 2678400) * pow(2, this.bits_per_dimensions));
                    break;
                }
                default: {
                    epoch_min = ChronoUnit.YEARS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_min);
                    epoch_max = ChronoUnit.YEARS.between(LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC), date_max);
                    LocalDateTime epoch_min_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusYears(epoch_min);
                    LocalDateTime epoch_max_date = LocalDateTime.ofEpochSecond(0, 0, ZoneOffset.UTC).plusYears(epoch_max);
                    long offset_min = ChronoUnit.MINUTES.between(epoch_min_date, date_min);
                    long offset_max = ChronoUnit.MINUTES.between(epoch_max_date, date_max);
                    scaled_offset_epoch_min = (long) (((double) offset_min / 527040) * pow(2, this.bits_per_dimensions));
                    scaled_offset_epoch_max = (long) (((double) offset_max / 527040) * pow(2, this.bits_per_dimensions));
                    break;
                }
            }
        } else {
            long offset_min = ChronoUnit.MINUTES.between(LocalDateTime.parse("2007-04-01 00:00:00", formatter), date_min);
            long offset_max = ChronoUnit.MINUTES.between(LocalDateTime.parse("2007-04-01 00:00:00", formatter), date_max);
            scaled_offset_epoch_min = (long) (((double) offset_min / 2720000) * pow(2, this.bits_per_dimensions)); // 350640 = hours in a fourty year interval
            scaled_offset_epoch_max = (long) (((double) offset_max / 2720000) * pow(2, this.bits_per_dimensions));
        }


        List<Coordinate> encoded_vertices = new ArrayList<>();
        List<Coordinate> vertices = new ArrayList<>();

        extractPolygonCoordinates(query.getSpatialPredicate(), encoded_vertices, vertices);


        Polygon encoded_poly = geometry_factory.createPolygon(encoded_vertices.toArray(new Coordinate[]{}));
        Polygon poly = geometry_factory.createPolygon(vertices.toArray(new Coordinate[]{}));


        List<Pair> candidate_ranges_and_epochs = new ArrayList();

        logger.endParsing();
        logger.startCreateRanges();

        if (this.enable_epoch) {
            long epoch_length = epoch_max - epoch_min + 1;
            query_planner.enableEpoch();
            for (int i = 0; i < epoch_length; i++) {

                QueryRegion bound;

                long epoch = epoch_min + i;

                if (epoch_length == 1 && i == 0) {
                    bound = new QueryRegion(this.bits_per_dimensions, encoded_poly, scaled_offset_epoch_min, scaled_offset_epoch_max, epoch);
                } else if (epoch_length > 1 && i == 0) {
                    bound = new QueryRegion(this.bits_per_dimensions, encoded_poly, scaled_offset_epoch_min, (long) pow(2, this.bits_per_dimensions) - 1, epoch);

                } else if (epoch_length > 1 && i == epoch_length - 1) {
                    bound = new QueryRegion(this.bits_per_dimensions, encoded_poly, 0, scaled_offset_epoch_max, epoch);
                } else {
                    bound = new QueryRegion(this.bits_per_dimensions, encoded_poly, 0, (long) pow(2, this.bits_per_dimensions) - 1, epoch);
                }

                List<ZorderRange> ranges = query_planner.regionDecompostion(bound, max_ranges, best_first_traversal);

                candidate_ranges_and_epochs.add(new Pair(ranges, epoch));
            }
        } else {

            QueryRegion bound = new QueryRegion(this.bits_per_dimensions, encoded_poly, scaled_offset_epoch_min, scaled_offset_epoch_max, 0);
            List<ZorderRange> ranges = query_planner.regionDecompostion(bound, max_ranges, best_first_traversal);
            candidate_ranges_and_epochs.add(new Pair(ranges, 0));
        }

        List<Collection<String>> features = new ArrayList<>();


        logger.endCreateRanges(candidate_ranges_and_epochs.get(0).candidate_ranges.size());
        logger.startExecution();


        if (this.multithreading) {
            this.service = Executors.newFixedThreadPool(this.cores);
        }


        List<Callable<RefinedFeatures>> tasks = new ArrayList<>();
        long false_positive_count = 0;


        for (Pair candidate_range_and_epoch : candidate_ranges_and_epochs) {
            for (ZorderRange range : candidate_range_and_epoch.candidate_ranges) {
                ConcurrentNavigableMap<long[], String> features_retrieved;
                if (enable_epoch) {
                    features_retrieved =
                            map.subMap(
                                    new long[]{candidate_range_and_epoch.epoch, range.lower_range}, true,
                                    new long[]{candidate_range_and_epoch.epoch, range.upper_range}, true
                            );
                } else {
                    features_retrieved =
                            map.subMap(
                                    new long[]{range.lower_range}, true,
                                    new long[]{range.upper_range}, true
                            );
                }


                if (range.is_contained) {

                    features.add(features_retrieved.values());


                } else {


                    if (this.multithreading) {

                        Filter filter = new Filter(features_retrieved);
                        filter.setSpatialPredicate(poly);
                        filter.setTemporalPredicate(date_min, date_max);
                        tasks.add(filter);

                    } else {

                        Filter filter = new Filter(features_retrieved);
                        filter.setSpatialPredicate(poly);
                        filter.setTemporalPredicate(date_min, date_max);
                        RefinedFeatures filtered_features = filter.call();
                        if (!filtered_features.getRefinedFeatures().isEmpty()) {
                            if (fetch_data) {
                                logger.startFetching();
                                features.add(filtered_features.getRefinedFeatures());
                                logger.endFetching();
                            }
                            false_positive_count += filtered_features.getFalsePositiveCount();
                        }


                    }

                }
            }
        }

        long[] false_positive_and_refined_count = multithreadFilter(features, tasks, fetch_data, logger);
        false_positive_count += false_positive_and_refined_count[0];

        Iterable<String> results = Iterables.concat(features);

        long count = 0;

        if (fetch_data) {
            logger.startFetching();
            count = fetchData(results);
            logger.endFetching();
        }

        logger.endExecution(count, false_positive_count);


        return results;
    }

    private Collection<String> addFeaturesToList(Collection<Object[]> refined_features) {

        Collection<String> features = new ArrayList<>();
        for (Object[] key : refined_features) {
            features.add(map.get(key));
        }
        return features;

    }

    private long[] multithreadFilter(List<Collection<String>> features, List<Callable<RefinedFeatures>> future_filtered_features_list, boolean fetch_data, Logger logger) {
        long false_positive_count = 0;
        long refined_count = 0;
        if (this.multithreading) {

            try {
                List<Future<RefinedFeatures>> futures = this.service.invokeAll(future_filtered_features_list);
                for (Future<RefinedFeatures> future : futures) {
                    if (!future.isCancelled()) {
                        RefinedFeatures filtered_features = future.get();
                        if (!filtered_features.getRefinedFeatures().isEmpty()) {
                            logger.startFetching();
                            features.add(filtered_features.getRefinedFeatures());
                            logger.endFetching();
                            false_positive_count += filtered_features.getFalsePositiveCount();
                            refined_count += filtered_features.getRefinedCount();
                        }

                    }
                }

            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
            this.service.shutdown();


        }
        return new long[]{false_positive_count, refined_count};
    }

    public void extractPolygonCoordinates(List<String[]> spatial_predicate, List<Coordinate> encoded_vertices, List<Coordinate> vertices) {
        for (String[] vertice : spatial_predicate) {

            long scaled_lat = ((long) ((Double.parseDouble(vertice[0]) + 90) / 180 * pow(2, this.bits_per_dimensions)));
            long scaled_lon = ((long) ((Double.parseDouble(vertice[1]) + 180) / 360 * pow(2, this.bits_per_dimensions)));
            encoded_vertices.add(new Coordinate(scaled_lat, scaled_lon));
            vertices.add(new Coordinate(Double.parseDouble(vertice[0]), Double.parseDouble(vertice[1])));

        }
    }

    @Benchmark
    private long fetchData(Iterable<String> features) {
        long counter = 0;
        for (String feat : features) {
            if (feat.length() != '0') { // Will never happen, avoids dead code elimination by JIT
                counter++;

            }
        }
        return counter;
    }

    public long countData(List<Collection<String>> features) {
        long counter = 0;
        for (Collection<String> feat : features) {
            counter += feat.size();

        }
        return counter;

    }


}
