package master;

import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.PriorityQueue;
import java.util.concurrent.*;


public class QueryPlanner {


    private boolean multithreading;
    private int cores;


    int dimensions;
    int bits_per_dimensions;
    FrequencyPrefixHistogram frequency_prefix_histogram;
    boolean enable_epoch;


    public QueryPlanner(int dimensions, int bits) {
        this.dimensions = dimensions;
        this.bits_per_dimensions = bits;
        this.enable_epoch = false;
    }

    public void enableMultithreading(int cores) {
        this.multithreading = true;
        this.cores = cores;
    }

    public void enableEpoch() {
        this.enable_epoch = true;
    }

    public void setFrequencyPrefixHistogram(FrequencyPrefixHistogram FrequencyPrefixHistogram) {
        this.frequency_prefix_histogram = FrequencyPrefixHistogram;
    }


    public List<ZorderRange> regionDecompostion(QueryRegion zbounds, int max_ranges, boolean best_first_traversal) {


        Decomposition initial_search = new Decomposition(this.dimensions, this.bits_per_dimensions, this.enable_epoch);


        long[] root_hyper_quadrant_prefix_and_length = initial_search.getLongestCommonPrefix(new long[]{zbounds.z_bound_lower, zbounds.z_bound_upper});
        long root_hyper_quadrant_prefix = root_hyper_quadrant_prefix_and_length[0];
        long root_hyper_quadrant_prefix_length = root_hyper_quadrant_prefix_and_length[1];
        long offset = 64 - root_hyper_quadrant_prefix_length;

        if (multithreading) {
            initial_search.setDecompositionParameters(zbounds, 100, frequency_prefix_histogram, 0, best_first_traversal, 0.01);
        } else {
            initial_search.setDecompositionParameters(zbounds, max_ranges, frequency_prefix_histogram, 0, best_first_traversal, 0.01);
        }
        initial_search.setOffset(offset);
        initial_search.checkHyperQuadrant(root_hyper_quadrant_prefix, 0, -1);
        offset -= this.dimensions;
        initial_search.setOffset(offset);
        List<ZorderRange> index_ranges = new ArrayList<>(100000);

        if (multithreading) {
            ExecutorService service = Executors.newFixedThreadPool(this.cores);
            index_ranges.addAll(initial_search.createRanges(true));
            List<UnprocessedHyperQuadrant> initial_remaining_list = new ArrayList<>(initial_search.getUnprocessedHyperQuadrants());
            List<Callable<List<ZorderRange>>> tasks = new ArrayList<>();

            for (int i = 0; i < this.cores; i++) {
                PriorityQueue<UnprocessedHyperQuadrant> priority_queue = new PriorityQueue<UnprocessedHyperQuadrant>(10);
                for (int j = 0; j < initial_remaining_list.size() - i; j += this.cores) {
                    priority_queue.add(initial_remaining_list.get(i + j));
                }
                Decomposition search_thread = new Decomposition(this.dimensions, this.bits_per_dimensions, this.enable_epoch);
                search_thread.setDecompositionParameters(zbounds, max_ranges / this.cores, frequency_prefix_histogram, i, best_first_traversal, 0.01);
                search_thread.setOffset(offset);
                search_thread.setUnprocessedHyperQuadrants(priority_queue);
                tasks.add(search_thread);
            }


            try {
                List<Future<List<ZorderRange>>> future_index_range = service.invokeAll(tasks);
                for (Future<List<ZorderRange>> future : future_index_range) {
                    if (!future.isCancelled()) {
                        List<ZorderRange> ranges = future.get();
                        for (ZorderRange r : ranges) {
                            index_ranges.add(r);
                        }
                    }
                }

            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
            service.shutdown();
        } else {
            index_ranges = initial_search.createRanges(false);
        }

        return getMergedIndexRanges(index_ranges);

    }


    @NotNull
    private List<ZorderRange> getMergedIndexRanges(List<ZorderRange> index_ranges) {
        List<ZorderRange> result = new ArrayList<>();

        if (!index_ranges.isEmpty()) {

            Collections.sort(index_ranges);

            ZorderRange current = index_ranges.get(0);

            for (int i = 1; i < index_ranges.size(); i++) {

                ZorderRange range = index_ranges.get(i);

                if ((range.lower_range == (current.upper_range + 1)) && (current.is_contained == range.is_contained)) {

                    current = new ZorderRange(current.lower_range, range.upper_range, current.is_contained);

                } else {
                    result.add(current);
                    current = range;
                }
            }
            result.add(current);
        }
        return result;
    }

}
